#!/usr/bin/python3
"""Checks that neighbor-list reallocation correctly takes place"""

from __future__ import absolute_import
from __future__ import print_function
import unittest
import sys

py_rumd_lib = {2:"../../../lib", 3:"../../../lib_py3"}[sys.version_info[0]]
sys.path.insert(1, py_rumd_lib)
sys.path.insert(1, "../../../Python")


import rumd
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)
import rumd.Simulation

class TestNBListAllocation(unittest.TestCase):
    """Check that the exclusion list and exclusion by type
    work as they should """
    def setUp(self):
        pass

    def test_calc_energy(self):
        """ Calculate potential energy by hand and compare to rumd CalcF, just
        to check basic operation before doing exclusions """

        sim = rumd.Simulation.Simulation("startN1000.xyz.gz", 32, 4,
                                         verbose=False)

        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.NoShift)
        # Set up parameters for Lennard-Jones interactions
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)

        potential.SetParams(0, 0, 1.00, 1.00, 2.5)
        potential.SetParams(0, 1, 0.8, 1.50, 2.5)
        potential.SetParams(1, 1, 0.88, 0.5, 2.5)


        sim.AddPotential(potential)
        potential.SetNB_Method("none")
        print("none (reference)")
        sim.sample.CalcF()
        pe_ref = sim.sample.GetPotentialEnergy()


        potential.SetNB_Method("n2")
        print("n2")

        sim.sample.SetPB_TP(32, 1)
        potential.ResetInternalData()
        sim.sample.CalcF()
        #pe_n2 = sim.sample.GetPotentialEnergy()
        actual_max_num_nbrs = potential.GetActualMaxNumNbrs()
        print("actual_max_num_nbrs", actual_max_num_nbrs)

        print("n2_tp1")
        sim.sample.SetPB_TP(32, 1)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_n2_tp1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("n2_tp4")
        sim.sample.SetPB_TP(32, 4)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_n2_tp4 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("actual_max_num_nbrs-1", actual_max_num_nbrs-1)

        print("n2_tp1_nm1")
        sim.sample.SetPB_TP(32, 1)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs - 1)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_n2_tp1_nm1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("n2_tp4_nm1")
        sim.sample.SetPB_TP(32, 4)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs - 1)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_n2_tp4_nm1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("sort")
        potential.SetNB_Method("sort")

        print("sort_tp1")
        sim.sample.SetPB_TP(32, 1)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_sort_tp1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("sort_tp4")
        sim.sample.SetPB_TP(32, 4)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_sort_tp4 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("sort_tp1_nm1")
        sim.sample.SetPB_TP(32, 1)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs - 1)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_sort_tp1_nm1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)

        print("sort_tp4_nm1")
        sim.sample.SetPB_TP(32, 4)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs - 1)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_sort_tp4_nm1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)


        # the following is to check the case of having allocated just enough
        # memory (which now means actual max plus one) and tp>1
        # cuda-memcheck should detect a problem if the if statement
        # in the calf tp>1 function is commented out.
        print("sort_tp4_np1")
        sim.sample.SetPB_TP(32, 4)
        potential.SetNbMaxNumNbrs(actual_max_num_nbrs + 1)
        potential.ResetInternalData()
        sim.sample.CalcF()
        pe_sort_tp4_np1 = sim.sample.GetPotentialEnergy()
        self.assertEqual(potential.GetActualMaxNumNbrs(), actual_max_num_nbrs)


        print("%.5f %.5f %.5f %.5f %.5f" % (pe_ref, pe_n2_tp1, pe_n2_tp4,
                                            pe_n2_tp1_nm1, pe_n2_tp4_nm1))

        print("\t%.5f %.5f %.5f %.5f" % (pe_sort_tp1, pe_sort_tp4,
                                         pe_sort_tp1_nm1, pe_sort_tp4_nm1))
        print("%.5f" % pe_sort_tp4_np1)


        self.assertLess(abs((pe_n2_tp1 - pe_ref)), 1.e-4)
        self.assertLess(abs((pe_n2_tp4 - pe_ref)), 1.e-4)
        self.assertLess(abs((pe_n2_tp1_nm1 - pe_ref)), 1.e-4)
        self.assertLess(abs((pe_n2_tp4_nm1 - pe_ref)), 1.e-4)

        self.assertLess(abs((pe_sort_tp1 - pe_ref)), 1.e-4)
        self.assertLess(abs((pe_sort_tp4 - pe_ref)), 1.e-4)
        self.assertLess(abs((pe_sort_tp1_nm1 - pe_ref)), 1.e-4)
        self.assertLess(abs((pe_sort_tp4_nm1 - pe_ref)), 1.e-4)

        self.assertLess(abs((pe_sort_tp4_np1 - pe_ref)), 1.e-4)




    def tearDown(self):
        pass

if __name__ == '__main__':
    unittest.main()
